import logging

from dbtmetabase.format import setup_logging
from tests.test_exposures import *
from tests.test_format import *
from tests.test_manifest import *
from tests.test_metabase import *
from tests.test_models import *

setup_logging(level=logging.DEBUG, path=None)



from __future__ import annotations

import dataclasses as dc
import logging
import re
from abc import ABCMeta, abstractmethod
from operator import itemgetter
from pathlib import Path
from typing import (
    Iterable,
    Mapping,
    MutableMapping,
    MutableSequence,
    Optional,
    Sequence,
    Tuple,
)

from dbtmetabase.metabase import Metabase

from .errors import ArgumentError
from .format import Filter, dump_yaml, safe_description, safe_name
from .manifest import Manifest

_RESOURCE_VERSION = 2

# Extracting table in `from` and `join` clauses (won't recognize some valid SQL, e.g. `from "table with spaces"`)
_EXPOSURE_PARSER = re.compile(r"[FfJj][RrOo][OoIi][MmNn]\s+([\w.\"`]+)")
_CTE_PARSER = re.compile(
    r"[Ww][Ii][Tt][Hh]\s+\b(\w+)\b\s+as|[)]\s*[,]\s*\b(\w+)\b\s+as"
)

_logger = logging.getLogger(__name__)


class ExposuresMixin(metaclass=ABCMeta):
    """Abstraction for extracting exposures."""

    DEFAULT_EXPOSURES_OUTPUT_PATH = "."

    @property
    @abstractmethod
    def manifest(self) -> Manifest:
        pass

    @property
    @abstractmethod
    def metabase(self) -> Metabase:
        pass

    def extract_exposures(
        self,
        output_path: str = DEFAULT_EXPOSURES_OUTPUT_PATH,
        output_grouping: Optional[str] = None,
        collection_filter: Optional[Filter] = None,
        allow_personal_collections: bool = False,
        exclude_unverified: bool = False,
        tags: Optional[Sequence[str]] = None,
    ) -> Iterable[Mapping]:
        """Extract dbt exposures from Metabase.

        Args:
            output_path (str, optional): Path for output files. Defaults to ".".
            output_grouping (Optional[str], optional): Grouping for output YAML files, supported values: "collection" (by collection slug) or "type" (by entity type). Defaults to None.
            collection_filter (Optional[Filter], optional): Filter Metabase collections. Defaults to None.
            allow_personal_collections (bool, optional): Allow personal Metabase collections. Defaults to False.
            exclude_unverified (bool, optional): Exclude items that have not been verified. Only applies to entity types that support verification. Defaults to False.
            tags (Sequence[str], optional): Optional tags for exported dbt exposures. Defaults to None.

        Returns:
            Iterable[Mapping]: List of parsed exposures.
        """

        if output_grouping not in (None, "collection", "type"):
            raise ArgumentError(f"Unsupported grouping: {output_grouping}")

        collection_filter = collection_filter or Filter()

        models = self.manifest.read_models()

        ctx = self.__Context(
            model_refs={m.alias.lower(): m.ref for m in models if m.ref},
            table_names={t["id"]: t["name"] for t in self.metabase.get_tables()},
        )

        exposures = []
        counts: MutableMapping[str, int] = {}

        for collection in self.metabase.get_collections(
            exclude_personal=not allow_personal_collections
        ):
            collection_name = collection["name"]
            collection_slug = collection.get("slug", safe_name(collection["name"]))

            if not collection_filter.match(collection_name):
                _logger.debug("Skipping collection '%s'", collection["name"])
                continue

            _logger.info("Exploring collection '%s'", collection["name"])
            for item in self.metabase.get_collection_items(
                uid=collection["id"],
                models=("card", "dashboard"),
            ):
                if (
                    exclude_unverified
                    and item["model"] == "card"
                    and item.get("moderated_status") != "verified"
                ):
                    _logger.debug("Skipping unverified card '%s'", item["name"])
                    continue

                depends = set()
                native_query = ""
                header = ""
                average_query_time = None
                last_used_at = None

                entity: Mapping
                if item["model"] == "card":
                    card_entity = self.metabase.find_card(uid=item["id"])
                    if card_entity is None:
                        _logger.info("Card '%s' not found, skipping", item["id"])
                        continue

                    entity = card_entity
                    header = (
                        f"Visualization: {entity.get('display', 'Unknown').title()}"
                    )

                    result = self.__extract_card_exposures(ctx, card=entity)
                    depends.update(result["depends"])
                    native_query = result["native_query"]

                    average_query_time_ms = entity.get("average_query_time")
                    if average_query_time_ms:
                        average_query_time_s = average_query_time_ms / 1000
                        average_query_time = f"{(average_query_time_s // 60):.0f}:{(average_query_time_s % 60):06.3f}"

                    last_used_at = entity.get("last_used_at")

                elif item["model"] == "dashboard":
                    dashboard_entity = self.metabase.find_dashboard(uid=item["id"])
                    if dashboard_entity is None:
                        _logger.info("Dashboard '%s' not found, skipping", item["id"])
                        continue

                    entity = dashboard_entity
                    cards = entity.get("dashcards", entity.get("ordered_cards", []))
                    if not cards:
                        continue

                    header = f"Dashboard Cards: {len(cards)}"
                    for card_ref in cards:
                        card = card_ref.get("card", {})
                        if "id" not in card:
                            continue

                        depends.update(
                            self.__extract_card_exposures(
                                ctx,
                                card=self.metabase.find_card(uid=card["id"]),
                            )["depends"]
                        )
                else:
                    _logger.warning("Unexpected collection item '%s'", item["model"])
                    continue

                name = entity.get("name", "Exposure [Unresolved Name]")
                _logger.info("Processing %s '%s'", item["model"], name)

                creator_name = None
                creator_email = None
                if "creator" in entity:
                    creator_email = entity["creator"]["email"]
                    creator_name = entity["creator"]["common_name"]
                elif "creator_id" in entity:
                    creator = self.metabase.find_user(uid=entity["creator_id"])
                    if creator:
                        creator_name = creator.get("common_name")
                        creator_email = creator.get("email")

                label = name
                name = safe_name(name)
                count = counts.get(name, 0)
                counts[name] = count + 1

                exposures.append(
                    {
                        "id": item["id"],
                        "type": item["model"],
                        "collection": collection_slug,
                        "body": self.__format_exposure(
                            model=item["model"],
                            uid=item["id"],
                            name=name + (f"_{count}" if count > 0 else ""),
                            label=label,
                            header=header,
                            description=entity.get("description", ""),
                            created_at=entity["created_at"],
                            creator_name=creator_name or "",
                            creator_email=creator_email or "",
                            last_used_at=last_used_at,
                            average_query_time=average_query_time,
                            native_query=native_query,
                            depends_on=sorted(
                                [
                                    ctx.model_refs[depend.lower()]
                                    for depend in depends
                                    if depend.lower() in ctx.model_refs
                                ]
                            ),
                            tags=tags,
                        ),
                    }
                )

        self.__write_exposures(exposures, output_path, output_grouping)

        return exposures

    def __extract_card_exposures(
        self,
        ctx: __Context,
        card: Optional[Mapping],
    ) -> Mapping:
        """Extracts exposures from Metabase questions."""

        depends = set()
        native_query = ""

        if card:
            query = card.get("dataset_query", {})
            if query.get("type") == "query":
                # Metabase GUI derived query
                query_source = query.get("query", {}).get(
                    "source-table", card.get("table_id")
                )

                if str(query_source).startswith("card__"):
                    # Handle questions based on other questions
                    depends.update(
                        self.__extract_card_exposures(
                            ctx,
                            card=self.metabase.find_card(
                                uid=query_source.split("__")[-1]
                            ),
                        )["depends"]
                    )
                elif query_source in ctx.table_names:
                    # Normal question
                    source_table = ctx.table_names.get(query_source)
                    if source_table:
                        source_table = source_table.lower()
                        _logger.info("Extracted model '%s' from card", source_table)
                        depends.add(source_table)

                # Find models exposed through joins
                for join in query.get("query", {}).get("joins", []):
                    join_source = join.get("source-table")

                    if str(join_source).startswith("card__"):
                        # Handle questions based on other questions
                        depends.update(
                            self.__extract_card_exposures(
                                ctx,
                                card=self.metabase.find_card(
                                    uid=join_source.split("__")[-1]
                                ),
                            )["depends"]
                        )
                        continue

                    # Joined model parsed
                    joined_table = ctx.table_names.get(join_source)
                    if joined_table:
                        joined_table = joined_table.lower()
                        _logger.info("Extracted model '%s' from join", joined_table)
                        depends.add(joined_table)

            elif query.get("type") == "native":
                # Metabase native query
                native_query = query["native"].get("query")
                ctes: MutableSequence[str] = []

                # Parse common table expressions for exclusion
                for matched_cte in re.findall(_CTE_PARSER, native_query):
                    ctes.extend(group.lower() for group in matched_cte if group)

                # Parse SQL for exposures through FROM or JOIN clauses
                for sql_ref in re.findall(_EXPOSURE_PARSER, native_query):
                    # Grab just the table / model name
                    parsed_model = sql_ref.split(".")[-1].strip('"').lower()

                    # Scrub CTEs (qualified sql_refs can not reference CTEs)
                    if parsed_model in ctes and "." not in sql_ref:
                        continue

                    # Verify this is one of our parsed refable models so exposures dont break the DAG
                    if not ctx.model_refs.get(parsed_model):
                        continue

                    if parsed_model:
                        _logger.info(
                            "Extracted model '%s' from native query", parsed_model
                        )
                        depends.add(parsed_model)

        return {
            "depends": depends,
            "native_query": native_query,
        }

    def __format_exposure(
        self,
        model: str,
        uid: str,
        name: str,
        label: str,
        header: str,
        description: str,
        created_at: str,
        creator_name: str,
        creator_email: str,
        last_used_at: Optional[str],
        average_query_time: Optional[str],
        native_query: Optional[str],
        depends_on: Iterable[str],
        tags: Optional[Sequence[str]],
    ) -> Mapping:
        """Builds dbt exposure representation (see https://docs.getdbt.com/reference/exposure-properties)."""

        dbt_type: str
        url: str
        if model == "card":
            dbt_type = "analysis"
            url = self.metabase.format_card_url(uid=uid)
        elif model == "dashboard":
            dbt_type = "dashboard"
            url = self.metabase.format_dashboard_url(uid=uid)
        else:
            raise ValueError("Unexpected exposure type")

        if header:
            header = f"### {header}\n\n"

        if description:
            description = description.strip()
        else:
            description = "No description provided in Metabase"

        if native_query:
            # Format query into markdown code block
            native_query = "\n".join(x for x in native_query.split("\n") if x.strip())
            native_query = f"#### Query\n\n```\n{native_query}\n```\n\n"

        metadata = (
            "#### Metadata\n\n"
            + f"Metabase ID: __{uid}__\n\n"
            + f"Created On: __{created_at}__"
        )

        exposure = {
            "name": name,
            "label": label,
            "description": safe_description(
                f"{header}{description}\n\n{native_query}{metadata}"
            ),
            "type": dbt_type,
            "url": url,
            "maturity": "medium",
            "owner": {
                "name": creator_name,
                "email": creator_email,
            },
            "depends_on": list(depends_on),
        }

        meta = {}
        if average_query_time:
            meta["average_query_time"] = average_query_time
        if last_used_at:
            meta["last_used_at"] = last_used_at
        if meta:
            exposure["meta"] = meta

        if tags:
            exposure["tags"] = list(tags)

        return exposure

    @staticmethod
    def __write_exposures(
        exposures: Iterable[Mapping],
        output_path: str,
        output_grouping: Optional[str],
    ):
        """Write exposures to output files."""

        grouped: MutableMapping[Tuple[str, ...], MutableSequence[Mapping]] = {}
        for exposure in exposures:
            group: Tuple[str, ...] = ("exposures",)
            if output_grouping == "collection":
                group = (exposure["collection"],)
            elif output_grouping == "type":
                group = (exposure["type"], exposure["id"])

            exps = grouped.get(group, [])
            exps.append(exposure)
            if group not in grouped:
                grouped[group] = exps

        for group, exps in grouped.items():
            path = Path(output_path).expanduser()
            path = path.joinpath(*group[:-1]) / f"{group[-1]}.yml"
            path.parent.mkdir(parents=True, exist_ok=True)

            exps_unwrapped = (x["body"] for x in exps)
            exps_sorted = sorted(exps_unwrapped, key=itemgetter("name"))

            with open(path, "w", encoding="utf-8") as f:
                dump_yaml(
                    data={
                        "version": _RESOURCE_VERSION,
                        "exposures": exps_sorted,
                    },
                    stream=f,
                )

    @dc.dataclass
    class __Context:
        model_refs: Mapping[str, str]
        table_names: Mapping[str, str]



#!/usr/bin/env python
import logging

import requests
from molot import envarg, envarg_int, evaluate, shell, target

POSTGRES_HOST = envarg("POSTGRES_HOST")
POSTGRES_PORT = envarg_int("POSTGRES_PORT")
POSTGRES_DB = envarg("POSTGRES_DB")
POSTGRES_USER = envarg("POSTGRES_USER")
POSTGRES_PASSWORD = envarg("POSTGRES_PASSWORD")
MB_HOST = envarg("MB_HOST")
MB_PORT = envarg_int("MB_PORT")
MB_SETUP_TOKEN = envarg("MB_SETUP_TOKEN")
MB_USER = envarg("MB_USER")
MB_PASSWORD = envarg("MB_PASSWORD")
MB_NAME = envarg("MB_NAME", "dbtmetabase")

MB_API_URL = f"http://{MB_HOST}:{MB_PORT}/api"


@target(
    description="initial setup",
    depends=["dbt_run", "metabase_setup"],
)
def init():
    pass


@target(description="run dbt project")
def dbt_run():
    shell("dbt seed --profiles-dir .")
    shell("dbt run --profiles-dir .")


@target(description="set up Metabase user and database")
def metabase_setup():
    setup_resp = requests.post(
        url=f"{MB_API_URL}/setup",
        json={
            "token": MB_SETUP_TOKEN,
            "user": {
                "site_name": MB_NAME,
                "first_name": MB_NAME,
                "last_name": None,
                "email": MB_USER,
                "password_confirm": MB_PASSWORD,
                "password": MB_PASSWORD,
            },
            "prefs": {
                "site_name": MB_NAME,
                "site_locale": "en",
                "allow_tracking": "false",
            },
        },
        timeout=10,
    )
    if setup_resp.status_code == 200:
        logging.info("Metabase setup successful")
    elif setup_resp.status_code == 403:
        logging.info("Metabase already set up")
    else:
        raise requests.HTTPError(f"Error: {setup_resp.reason}", response=setup_resp)

    session_id = requests.post(
        url=f"{MB_API_URL}/session",
        json={"username": MB_USER, "password": MB_PASSWORD},
        timeout=10,
    ).json()["id"]

    headers = {"X-Metabase-Session": session_id}

    database_id = ""
    sample_database_id = ""
    databases = requests.get(
        url=f"{MB_API_URL}/database",
        headers=headers,
        timeout=10,
    ).json()["data"]
    for db in databases:
        if db["name"] == POSTGRES_DB and db["engine"] == "postgres":
            database_id = db["id"]
        elif db["name"] == "Sample Database" and db["engine"] == "h2":
            sample_database_id = db["id"]

    if sample_database_id:
        logging.info("Archiving Metabase sample database %s", sample_database_id)
        requests.delete(
            url=f"{MB_API_URL}/database/{sample_database_id}",
            headers=headers,
            timeout=10,
        ).raise_for_status()

    collections = requests.get(
        url=f"{MB_API_URL}/collection",
        headers=headers,
        timeout=10,
    ).json()
    for collection in collections:
        if collection.get("is_sample") and not collection.get("archived"):
            logging.info("Deleting Metabase sample collection %s", collection["id"])
            requests.put(
                url=f"{MB_API_URL}/collection/{collection['id']}",
                headers=headers,
                json={"archived": True},
                timeout=10,
            ).raise_for_status()

    database_body = {
        "engine": "postgres",
        "name": POSTGRES_DB,
        "details": {
            "host": POSTGRES_HOST,
            "port": POSTGRES_PORT,
            "dbname": POSTGRES_DB,
            "user": POSTGRES_USER,
            "password": POSTGRES_PASSWORD,
            "schema-filters-type": "all",
            "ssl": False,
            "tunnel-enabled": False,
            "advanced-options": False,
        },
        "is_on_demand": False,
        "is_full_sync": True,
        "is_sample": False,
        "cache_ttl": None,
        "refingerprint": False,
        "auto_run_queries": True,
        "schedules": {},
    }
    if not database_id:
        logging.info("Creating Metabase database")
        database_id = requests.post(
            url=f"{MB_API_URL}/database",
            headers=headers,
            json=database_body,
            timeout=10,
        ).json()["id"]
    else:
        logging.info("Updating Metabase database %s", database_id)
        requests.put(
            url=f"{MB_API_URL}/database/{database_id}",
            headers=headers,
            json=database_body,
            timeout=10,
        ).raise_for_status()

    logging.info("Triggering Metabase database sync")
    requests.post(
        url=f"{MB_API_URL}/database/{database_id}/sync_schema",
        headers=headers,
        json={},
        timeout=10,
    ).raise_for_status()


evaluate()



import json
import os
from pathlib import Path
from typing import Any, Dict, Mapping, Optional, Sequence, Union

import requests
from dotenv import dotenv_values

from dbtmetabase.core import DbtMetabase
from dbtmetabase.manifest import Column, Manifest, Model
from dbtmetabase.metabase import Metabase

FIXTURES_PATH = Path("tests") / "fixtures"
TMP_PATH = Path("tests") / "tmp"

RECORD = os.getenv("RECORD", "").lower() == "true"
SANDBOX_ENV = dotenv_values(Path().parent / "sandbox" / ".env")


class MockMetabase(Metabase):
    def __init__(self, url: str, record: bool = False):
        self.record = record

        api_key = "dummy"
        username = None
        password = None

        if record:
            api_key = None
            username = SANDBOX_ENV["MB_USER"]
            password = SANDBOX_ENV["MB_PASSWORD"]

        super().__init__(
            url=url,
            api_key=api_key,
            username=username,
            password=password,
            session_id=None,
            skip_verify=False,
            cert=None,
            http_timeout=1,
            http_headers=None,
            http_adapter=None,
        )

    def _api(
        self,
        method: str,
        path: str,
        params: Optional[Dict[str, Any]] = None,
        **kwargs,
    ) -> Union[Mapping, Sequence]:
        result = {}
        path_toks = f"{path.lstrip('/')}.json".split("/")
        json_path = Path.joinpath(FIXTURES_PATH, *path_toks)

        if self.record:
            is_auth = path == "/api/session"
            if method == "get" or is_auth:
                result = super()._api(method, path, params, **kwargs)

                if not is_auth:
                    json_path.parent.mkdir(parents=True, exist_ok=True)
                    with open(json_path, "w", encoding="utf-8") as f:
                        json.dump(result, f, indent=4)
        else:
            if method == "get":
                if json_path.exists():
                    with open(json_path, encoding="utf-8") as f:
                        result = json.load(f)
                else:
                    response = requests.Response()
                    response.status_code = 404
                    raise requests.exceptions.HTTPError(response=response)

        return result


class MockManifest(Manifest):
    _models: Sequence[Model] = []

    def read_models(self) -> Sequence[Model]:
        if not self._models:
            self._models = super().read_models()
        return self._models

    def find_model(self, model_name: str) -> Optional[Model]:
        filtered = [m for m in self._models if m.name == model_name]
        if filtered:
            return filtered[0]
        return None

    def find_column(
        self,
        model_name: str,
        column_name: str,
    ) -> Optional[Column]:
        model = self.find_model(model_name=model_name)
        if model:
            filtered = [c for c in model.columns if c.name == column_name]
            if filtered:
                return filtered[0]
        return None


class MockDbtMetabase(DbtMetabase):
    def __init__(
        self,
        manifest_path: Path = FIXTURES_PATH / "manifest-v12.json",
        metabase_url: str = f"http://localhost:{SANDBOX_ENV['MB_PORT']}",
    ):
        self._manifest = MockManifest(path=manifest_path)
        self._metabase = MockMetabase(url=metabase_url, record=RECORD)



import pytest

from tests._mocks import MockMetabase


@pytest.fixture(name="metabase")
def fixture_metabase() -> MockMetabase:
    return MockMetabase(url="http://localhost")


def test_metabase_find_database(metabase: MockMetabase):
    db = metabase.find_database(name="dbtmetabase")
    assert db
    assert db["id"] == 2
    assert metabase.find_database(name="foo") is None


def test_metabase_get_collections(metabase: MockMetabase):
    excluded = metabase.get_collections(exclude_personal=True)
    assert len(excluded) == 1

    included = metabase.get_collections(exclude_personal=False)
    assert len(included) == 2


def test_metabase_get_collection_items(metabase: MockMetabase):
    cards = metabase.get_collection_items(
        uid="root",
        models=("card",),
    )
    assert {item["model"] for item in cards} == {"card"}

    dashboards = metabase.get_collection_items(
        uid="root",
        models=("dashboard",),
    )
    assert {item["model"] for item in dashboards} == {"dashboard"}

    both = metabase.get_collection_items(
        uid="root",
        models=("card", "dashboard"),
    )
    assert {item["model"] for item in both} == {"card", "dashboard"}



from operator import itemgetter
from pathlib import Path

import pytest
import yaml

from tests._mocks import FIXTURES_PATH, TMP_PATH, MockDbtMetabase


def setup_module():
    TMP_PATH.mkdir(exist_ok=True)


@pytest.fixture(name="core")
def fixture_core() -> MockDbtMetabase:
    return MockDbtMetabase()


def _assert_exposures(expected_path: Path, actual_path: Path):
    with open(expected_path, encoding="utf-8") as f:
        expected = yaml.safe_load(f)
    with open(actual_path, encoding="utf-8") as f:
        actual = yaml.safe_load(f)

    assert actual["exposures"] == sorted(expected["exposures"], key=itemgetter("name"))


def test_exposures_default(core: MockDbtMetabase):
    fixtures_path = FIXTURES_PATH / "exposure" / "default"
    output_path = TMP_PATH / "exposure" / "default"
    core.extract_exposures(
        output_path=str(output_path),
        output_grouping=None,
        tags=["metabase"],
    )

    _assert_exposures(
        fixtures_path / "exposures.yml",
        output_path / "exposures.yml",
    )


def test_exposures_collection_grouping(core: MockDbtMetabase):
    fixtures_path = FIXTURES_PATH / "exposure" / "collection"
    output_path = TMP_PATH / "exposure" / "collection"
    core.extract_exposures(
        output_path=str(output_path),
        output_grouping="collection",
    )

    for file in fixtures_path.iterdir():
        _assert_exposures(file, output_path / file.name)


def test_exposures_grouping_type(core: MockDbtMetabase):
    fixtures_path = FIXTURES_PATH / "exposure" / "type"
    output_path = TMP_PATH / "exposure" / "type"
    core.extract_exposures(
        output_path=str(output_path),
        output_grouping="type",
    )

    for file in (fixtures_path / "card").iterdir():
        _assert_exposures(file, output_path / "card" / file.name)

    for file in (fixtures_path / "dashboard").iterdir():
        _assert_exposures(file, output_path / "dashboard" / file.name)



from dbtmetabase.format import Filter, NullValue, dump_yaml, safe_description, safe_name
from tests._mocks import FIXTURES_PATH, TMP_PATH


def test_filter():
    assert Filter(include=("alpHa", "bRavo")).match("Alpha")
    assert Filter().match("Alpha")
    assert Filter().match("")
    assert not Filter(include=("alpHa", "bRavo"), exclude=("alpha",)).match("Alpha")
    assert not Filter(exclude=("alpha",)).match("Alpha")
    assert Filter(include="alpha").match("Alpha")
    assert not Filter(exclude="alpha").match("Alpha")


def test_filter_wildcard():
    assert Filter(include="stg_*").match("stg_orders")
    assert Filter(include="STG_*").match("stg_ORDERS")
    assert not Filter(include="stg_*").match("orders")
    assert Filter(include="order?").match("orders")
    assert not Filter(include="order?").match("ordersz")
    assert Filter(include="*orders", exclude="stg_*").match("_orders")
    assert not Filter(include="*orders", exclude="stg_*").match("stg_orders")


def test_null_value():
    assert NullValue is not None
    assert not NullValue
    assert NullValue is NullValue


def test_safe_name():
    assert safe_name("Somebody's 2 collections!") == "somebody_s_2_collections_"
    assert safe_name("somebody_s_2_collections_") == "somebody_s_2_collections_"
    assert safe_name("") == ""


def test_safe_description():
    assert (
        safe_description("Depends on\n\nQuestion {{ #2 }}!")
        == "Depends on\n\nQuestion ( #2 )!"
    )
    assert (
        safe_description("Depends on\n\nQuestion ( #2 )!")
        == "Depends on\n\nQuestion ( #2 )!"
    )
    assert (
        safe_description("Depends on\n\nQuestion { #2 }!")
        == "Depends on\n\nQuestion { #2 }!"
    )
    assert (
        safe_description("{{start_date}} - cast({{rolling_days}})")
        == "(start_date) - cast((rolling_days))"
    )


def test_dump_yaml():
    fixture_path = FIXTURES_PATH / "test_dump_yaml.yml"
    output_path = TMP_PATH / "test_dump_yaml.yml"
    with open(output_path, "w", encoding="utf-8") as f:
        dump_yaml(
            data={
                "root": {
                    "attr1": "val1\nend",
                    "attr2": ["val2", "val3"],
                },
            },
            stream=f,
        )
    with open(output_path, "r", encoding="utf-8") as f:
        actual = f.read()
    with open(fixture_path, "r", encoding="utf-8") as f:
        expected = f.read()
    assert actual == expected



from operator import attrgetter
from typing import Sequence

from dbtmetabase.manifest import Column, Group, Manifest, Model
from tests._mocks import FIXTURES_PATH, MockManifest


def test_v11_disabled():
    manifest = MockManifest(FIXTURES_PATH / "manifest-v11-disabled.json")
    manifest.read_models()

    orders_mod = manifest.find_model("orders")
    assert orders_mod is None

    customer_id_col = manifest.find_column("customers", "customer_id")
    assert customer_id_col is not None
    assert customer_id_col.fk_target_table is None
    assert customer_id_col.fk_target_field is None


def test_v12():
    models = Manifest(FIXTURES_PATH / "manifest-v12.json").read_models()
    _assert_models_equal(
        models,
        [
            Model(
                database="dbtmetabase",
                schema="public",
                group=Group.nodes,
                name="payments",
                alias="transactions",
                description="This table has basic information about payments",
                unique_id="model.sandbox.payments",
                columns=[
                    Column(
                        name="payment_id",
                        description="This is a unique identifier for a payment",
                        semantic_type="type/PK",
                    ),
                    Column(
                        name="payment_method",
                        description="",
                    ),
                    Column(
                        name="order_id",
                        description="Foreign key to the orders table",
                        semantic_type="type/FK",
                        fk_target_table="public.orders",
                        fk_target_field="order_id",
                    ),
                    Column(
                        name="amount",
                        description="",
                    ),
                ],
            ),
            Model(
                database="dbtmetabase",
                schema="public",
                group=Group.nodes,
                name="customers",
                alias="customers",
                description="This table has basic information about a customer, as well as some derived facts based on a customer's orders",
                display_name="clients",
                unique_id="model.sandbox.customers",
                columns=[
                    Column(
                        name="customer_id",
                        description="This is a unique identifier for a customer",
                    ),
                    Column(
                        name="first_name",
                        description="Customer's first name. PII.",
                    ),
                    Column(
                        name="last_name",
                        description="Customer's last name. PII.",
                    ),
                    Column(
                        name="first_order",
                        description="Date (UTC) of a customer's first order",
                    ),
                    Column(
                        name="most_recent_order",
                        description="Date (UTC) of a customer's most recent order",
                    ),
                    Column(
                        name="number_of_orders",
                        description="Count of the number of orders a customer has placed",
                        display_name="order_count",
                    ),
                    Column(
                        name="customer_lifetime_value",
                        description="Total value (AUD) of a customer's orders",
                    ),
                ],
            ),
            Model(
                database="dbtmetabase",
                schema="public",
                group=Group.nodes,
                name="orders",
                alias="orders",
                description="This table has basic information about orders, as well as some derived facts based on payments",
                points_of_interest="Basic information only",
                caveats="Some facts are derived from payments",
                unique_id="model.sandbox.orders",
                columns=[
                    Column(
                        name="order_id",
                        description="This is a unique identifier for an order",
                        semantic_type="type/PK",
                    ),
                    Column(
                        name="customer_id",
                        description="Foreign key to the customers table",
                        semantic_type="type/FK",
                        fk_target_table="public.customers",
                        fk_target_field="customer_id",
                    ),
                    Column(
                        name="order_date",
                        description="Date (UTC) that the order was placed",
                    ),
                    Column(
                        name="status",
                        description="",
                    ),
                    Column(
                        name="amount",
                        description="Total amount (AUD) of the order",
                    ),
                    Column(
                        name="credit_card_amount",
                        description="Amount of the order (AUD) paid for by credit card",
                    ),
                    Column(
                        name="coupon_amount",
                        description="Amount of the order (AUD) paid for by coupon",
                    ),
                    Column(
                        name="bank_transfer_amount",
                        description="Amount of the order (AUD) paid for by bank transfer",
                    ),
                    Column(
                        name="gift_card_amount",
                        description="Amount of the order (AUD) paid for by gift card",
                    ),
                ],
            ),
            Model(
                database="dbtmetabase",
                schema="public",
                group=Group.nodes,
                name="stg_customers",
                alias="stg_customers",
                description="",
                unique_id="model.sandbox.stg_customers",
                visibility_type="hidden",
                columns=[
                    Column(
                        name="customer_id",
                        description="",
                    ),
                    Column(
                        name="first_name",
                        description="",
                    ),
                    Column(
                        name="last_name",
                        description="",
                    ),
                ],
            ),
            Model(
                database="dbtmetabase",
                schema="public",
                group=Group.nodes,
                name="stg_payments",
                alias="stg_payments",
                description="",
                unique_id="model.sandbox.stg_payments",
                columns=[
                    Column(
                        name="payment_id",
                        description="",
                    ),
                    Column(
                        name="payment_method",
                        description="",
                    ),
                    Column(
                        name="order_id",
                        description="",
                    ),
                    Column(
                        name="amount",
                        description="",
                    ),
                ],
            ),
            Model(
                database="dbtmetabase",
                schema="public",
                group=Group.nodes,
                name="stg_orders",
                alias="stg_orders",
                description="",
                unique_id="model.sandbox.stg_orders",
                columns=[
                    Column(
                        name="order_id",
                        description="",
                    ),
                    Column(
                        name="status",
                        description="",
                    ),
                    Column(
                        name="order_date",
                        description="",
                    ),
                    Column(
                        name="customer_id",
                        description="",
                    ),
                    Column(
                        name="sku_id",
                        description="",
                        semantic_type="type/FK",
                        fk_target_table="inventory.skus",
                        fk_target_field="sku_id",
                    ),
                ],
            ),
            Model(
                database="dbtmetabase",
                schema="inventory",
                group=Group.sources,
                name="skus",
                alias="skus",
                description="",
                unique_id="source.sandbox.inventory.skus",
                source="inventory",
            ),
        ],
    )


def test_v2():
    models = Manifest(FIXTURES_PATH / "manifest-v2.json").read_models()
    _assert_models_equal(
        models,
        [
            Model(
                database="test",
                schema="public",
                group=Group.nodes,
                name="orders",
                alias="orders",
                description="This table has basic information about orders, as well as some derived facts based on payments",
                unique_id="model.jaffle_shop.orders",
                columns=[
                    Column(
                        name="order_id",
                        description="This is a unique identifier for an order",
                    ),
                    Column(
                        name="customer_id",
                        description="Foreign key to the customers table",
                        semantic_type="type/FK",
                        fk_target_table="public.customers",
                        fk_target_field="customer_id",
                    ),
                    Column(
                        name="order_date",
                        description="Date (UTC) that the order was placed",
                    ),
                    Column(
                        name="status",
                        description="Orders can be one of the following statuses:\n\n| status         | description                                                                                                            |\n|----------------|------------------------------------------------------------------------------------------------------------------------|\n| placed         | The order has been placed but has not yet left the warehouse                                                           |\n| shipped        | The order has ben shipped to the customer and is currently in transit                                                  |\n| completed      | The order has been received by the customer                                                                            |\n| return_pending | The customer has indicated that they would like to return the order, but it has not yet been received at the warehouse |\n| returned       | The order has been returned by the customer and received at the warehouse                                              |",
                    ),
                    Column(
                        name="amount",
                        description="Total amount (AUD) of the order",
                    ),
                    Column(
                        name="credit_card_amount",
                        description="Amount of the order (AUD) paid for by credit card",
                    ),
                    Column(
                        name="coupon_amount",
                        description="Amount of the order (AUD) paid for by coupon",
                    ),
                    Column(
                        name="bank_transfer_amount",
                        description="Amount of the order (AUD) paid for by bank transfer",
                    ),
                    Column(
                        name="gift_card_amount",
                        description="Amount of the order (AUD) paid for by gift card",
                    ),
                ],
            ),
            Model(
                database="test",
                schema="public",
                group=Group.nodes,
                name="customers",
                alias="customers",
                description="This table has basic information about a customer, as well as some derived facts based on a customer's orders",
                unique_id="model.jaffle_shop.customers",
                columns=[
                    Column(
                        name="customer_id",
                        description="This is a unique identifier for a customer",
                        semantic_type=None,  # This is a PK field, should not be detected as FK
                    ),
                    Column(
                        name="first_name",
                        description="Customer's first name. PII.",
                    ),
                    Column(
                        name="last_name",
                        description="Customer's last name. PII.",
                    ),
                    Column(
                        name="first_order",
                        description="Date (UTC) of a customer's first order",
                    ),
                    Column(
                        name="most_recent_order",
                        description="Date (UTC) of a customer's most recent order",
                    ),
                    Column(
                        name="number_of_orders",
                        description="Count of the number of orders a customer has placed",
                    ),
                    Column(
                        name="customer_lifetime_value",
                        description="Total value (AUD) of a customer's orders",
                    ),
                ],
            ),
            Model(
                database="test",
                schema="public",
                group=Group.nodes,
                name="stg_orders",
                alias="stg_orders",
                description="",
                unique_id="model.jaffle_shop.stg_orders",
                columns=[
                    Column(
                        name="order_id",
                        description="",
                    ),
                    Column(
                        name="status",
                        description="",
                    ),
                ],
            ),
            Model(
                database="test",
                schema="public",
                group=Group.nodes,
                name="stg_payments",
                alias="stg_payments",
                description="",
                unique_id="model.jaffle_shop.stg_payments",
                columns=[
                    Column(
                        name="payment_id",
                        description="",
                    ),
                    Column(
                        name="payment_method",
                        description="",
                    ),
                ],
            ),
            Model(
                database="test",
                schema="public",
                group=Group.nodes,
                name="stg_customers",
                alias="stg_customers",
                description="",
                unique_id="model.jaffle_shop.stg_customers",
                tags=[],
                columns=[
                    Column(
                        name="customer_id",
                        description="",
                    )
                ],
            ),
        ],
    )


def _assert_models_equal(
    first: Sequence[Model],
    second: Sequence[Model],
):
    assert len(first) == len(second), "mismatched model count"

    first = sorted(first, key=attrgetter("name"))
    second = sorted(second, key=attrgetter("name"))

    for i, first_model in enumerate(first):
        second_model = second[i]
        assert first_model.name == second_model.name, "wrong model"
        assert len(first_model.columns) == len(
            second_model.columns
        ), f"mismatched column count in {first_model.name}"
        for j, first_column in enumerate(first_model.columns):
            second_column = second_model.columns[j]
            assert (
                first_column.name == second_column.name
            ), f"wrong column in model {first_model.name}"
            assert (
                first_column == second_column
            ), f"mismatched column {first_model.name}.{first_column.name}"
        assert first_model == second_model, f"mismatched model {first_model.name}"

    assert first == second



from typing import MutableSequence, cast

import pytest

from dbtmetabase.manifest import Column
from tests._mocks import MockDbtMetabase


@pytest.fixture(name="core")
def fixture_core() -> MockDbtMetabase:
    c = MockDbtMetabase()
    c._ModelsMixin__SYNC_PERIOD = 1  # type: ignore
    return c


def test_export(core: MockDbtMetabase):
    core.export_models(
        metabase_database="dbtmetabase",
        skip_sources=True,
        sync_timeout=1,
        order_fields=True,
    )


def test_export_hidden_table(core: MockDbtMetabase):
    core._manifest.read_models()
    model = core._manifest.find_model("stg_customers")
    assert model is not None
    model.visibility_type = "hidden"

    column = model.columns[0]
    column.name = "new_column_since_stale"
    columns = cast(MutableSequence[Column], model.columns)
    columns.append(column)

    core.export_models(
        metabase_database="dbtmetabase",
        skip_sources=True,
        sync_timeout=1,
        order_fields=True,
    )


def test_build_lookups(core: MockDbtMetabase):
    expected = {
        "PUBLIC.CUSTOMERS": {
            "CUSTOMER_ID",
            "FIRST_NAME",
            "LAST_NAME",
            "FIRST_ORDER",
            "MOST_RECENT_ORDER",
            "NUMBER_OF_ORDERS",
            "CUSTOMER_LIFETIME_VALUE",
        },
        "PUBLIC.ORDERS": {
            "ORDER_ID",
            "CUSTOMER_ID",
            "ORDER_DATE",
            "STATUS",
            "AMOUNT",
            "CREDIT_CARD_AMOUNT",
            "COUPON_AMOUNT",
            "BANK_TRANSFER_AMOUNT",
            "GIFT_CARD_AMOUNT",
        },
        "PUBLIC.TRANSACTIONS": {
            "PAYMENT_ID",
            "PAYMENT_METHOD",
            "ORDER_ID",
            "AMOUNT",
        },
        "PUBLIC.RAW_CUSTOMERS": {"ID", "FIRST_NAME", "LAST_NAME"},
        "PUBLIC.RAW_ORDERS": {"ID", "USER_ID", "ORDER_DATE", "STATUS"},
        "PUBLIC.RAW_PAYMENTS": {"ID", "ORDER_ID", "PAYMENT_METHOD", "AMOUNT"},
        "PUBLIC.STG_CUSTOMERS": {"CUSTOMER_ID", "FIRST_NAME", "LAST_NAME"},
        "PUBLIC.STG_ORDERS": {
            "ORDER_ID",
            "STATUS",
            "ORDER_DATE",
            "CUSTOMER_ID",
            "SKU_ID",
        },
        "PUBLIC.STG_PAYMENTS": {
            "PAYMENT_ID",
            "PAYMENT_METHOD",
            "ORDER_ID",
            "AMOUNT",
        },
        "INVENTORY.SKUS": {"SKU_ID", "PRODUCT"},
    }

    actual_tables = core._ModelsMixin__get_tables(database_id="2")  # type: ignore

    assert set(actual_tables.keys()) == set(expected.keys())

    for table, columns in expected.items():
        assert set(actual_tables[table]["fields"].keys()) == columns, f"table: {table}"
